/*
 *  Copyright 2012 Christopher Pheby
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.jadira.cdt.country;

import java.util.HashMap;
import java.util.Map;

import com.google.i18n.phonenumbers.PhoneNumberUtil;

import org.jadira.cdt.country.CountryCode;

/**
* Enumeration of ISO <a href="http://en.wikipedia.org/wiki/ISO_3166-1">ISO 3166-1</a>
* country codes.
* <p>
* The Enum names of this enumeration themselves are represented by
* <a href="http://en.wikipedia.org/wiki/ISO_3166-1_alpha-2">ISO 3166-1 alpha-2</a>
* codes. 
* Instance methods return country name ({@link ISOCountryCode#getName()}), 
* <a href="http://en.wikipedia.org/wiki/ISO_3166-1_alpha-3" >ISO 3166-1 alpha-3</a> codes ({@link ISOCountryCode#getAlpha3Code()}),
* <a href="http://en.wikipedia.org/wiki/ISO_3166-1_numericCode">ISO 3166-1 numericCode</a> codes ({@link ISOCountryCode#getNumericCode()}),
* <a href="http://en.wikipedia.org/wiki/E.164">and E.164 the dialling region</a> code ({@link ISOCountryCode#getDiallingRegionCode()})
* Static factory methods can retrieve ISOCountryCode instances that
* corresponds to a given code.
* </p>
* <p>
* To update this class, populate using countrynames.txt, downloadable from <a href="http://opengeocode.org/download/countrynames.txt"/>.
* For the purpose of FMPP, all but one line of header must be removed. 
* 
*/
public enum ISOCountryCode implements CountryCode {

    AF("Afghanistan", "AFG", 4),
    AX("Aland Islands", "ALA", 248),
    AL("Albania", "ALB", 8),
    DZ("Algeria", "DZA", 12),
    AS("American Samoa", "ASM", 16),
    AD("Andorra", "AND", 20),
    AO("Angola", "AGO", 24),
    AI("Anguilla", "AIA", 660),
    AQ("Antarctica", "ATA", 10),
    AG("Antigua and Barbuda", "ATG", 28),
    AR("Argentina", "ARG", 32),
    AM("Armenia", "ARM", 51),
    AW("Aruba", "ABW", 533),
    AU("Australia", "AUS", 36),
    AT("Austria", "AUT", 40),
    AZ("Azerbaijan", "AZE", 31),
    BS("Bahamas", "BHS", 44),
    BH("Bahrain", "BHR", 48),
    BD("Bangladesh", "BGD", 50),
    BB("Barbados", "BRB", 52),
    BY("Belarus", "BLR", 112),
    BE("Belgium", "BEL", 56),
    BZ("Belize", "BLZ", 84),
    BJ("Benin", "BEN", 204),
    BM("Bermuda", "BMU", 60),
    BT("Bhutan", "BTN", 64),
    BO("Plurinational State of Bolivia", "BOL", 68),
    BQ("Bonaire, Sint Eustatius and Saba", "BES", 535),
    BA("Bosnia and Herzegovina", "BIH", 70),
    BW("Botswana", "BWA", 72),
    BV("Bouvet Island", "BVT", 74),
    BR("Brazil", "BRA", 76),
    IO("British Indian Ocean Territory", "IOT", 86),
    BN("Brunei Darussalam", "BRN", 96),
    BG("Bulgaria", "BGR", 100),
    BF("Burkina Faso", "BFA", 854),
    BI("Burundi", "BDI", 108),
    KH("Cambodia", "KHM", 116),
    CM("Cameroon", "CMR", 120),
    CA("Canada", "CAN", 124),
    CV("Cape Verde", "CPV", 132),
    KY("Cayman Islands", "CYM", 136),
    CF("Central African Republic", "CAF", 140),
    TD("Chad", "TCD", 148),
    CL("Chile", "CHL", 152),
    CN("China", "CHN", 156),
    CX("Christmas Island", "CXR", 162),
    CC("Cocos (Keeling) Islands", "CCK", 166),
    CO("Colombia", "COL", 170),
    KM("Comoros", "COM", 174),
    CG("Congo", "COG", 178),
    CD("The Democratic Republic of the Congo", "COD", 180),
    CK("Cook Islands", "COK", 184),
    CR("Costa Rica", "CRI", 188),
    CI("Cote d'Ivoire", "CIV", 384),
    HR("Croatia", "HRV", 191),
    CU("Cuba", "CUB", 192),
    CW("Curacao", "CUW", 531),
    CY("Cyprus", "CYP", 196),
    CZ("Czech Republic", "CZE", 203),
    DK("Denmark", "DNK", 208),
    DJ("Djibouti", "DJI", 262),
    DM("Dominica", "DMA", 212),
    DO("Dominican Republic", "DOM", 214),
    EC("Ecuador", "ECU", 218),
    EG("Egypt", "EGY", 818),
    SV("El Salvador", "SLV", 222),
    GQ("Equatorial Guinea", "GNQ", 226),
    ER("Eritrea", "ERI", 232),
    EE("Estonia", "EST", 233),
    ET("Ethiopia", "ETH", 231),
    FK("Falkland Islands (Malvinas)", "FLK", 238),
    FO("Faroe Islands", "534", 703),
    FJ("Fiji", "FJI", 242),
    FI("Finland", "FIN", 246),
    FR("France", "FRA", 250),
    GF("French Guiana", "GUF", 254),
    PF("French Polynesia", "PYF", 258),
    TF("French Southern Territories", "ATF", 260),
    GA("Gabon", "GAB", 266),
    GM("Gambia", "GMB", 270),
    GE("Georgia", "GEO", 268),
    DE("Germany", "DEU", 276),
    GH("Ghana", "GHA", 288),
    GI("Gibraltar", "GIB", 292),
    GR("Greece", "GRC", 300),
    GL("Greenland", "GRL", 304),
    GD("Grenada", "GRD", 308),
    GP("Guadeloupe", "GLP", 312),
    GU("Guam", "GUM", 316),
    GT("Guatemala", "GTM", 320),
    GG("Guernsey", "GGY", 831),
    GN("Guinea", "GIN", 324),
    GW("Guinea-Bissau", "GNB", 624),
    GY("Guyana", "GUY", 328),
    HT("Haiti", "HTI", 332),
    HM("Heard Island and McDonald Islands", "HMD", 334),
    VA("Holy See (Vatican City State)", "VAT", 336),
    HN("Honduras", "HND", 340),
    HK("Hong Kong", "HKG", 344),
    HU("Hungary", "HUN", 348),
    IS("Iceland", "ISL", 352),
    IN("India", "IND", 356),
    ID("Indonesia", "IDN", 360),
    IR("Islamic Republic of Iran", "IRN", 364),
    IQ("Iraq", "IRQ", 368),
    IE("Ireland", "IRL", 372),
    IM("Isle of Man", "IMN", 833),
    IL("Israel", "ISR", 376),
    IT("Italy", "ITA", 380),
    JM("Jamaica", "JAM", 388),
    JP("Japan", "JPN", 392),
    JE("Jersey", "JEY", 832),
    JO("Jordan", "JOR", 400),
    KZ("Kazakhstan", "KAZ", 398),
    KE("Kenya", "KEN", 404),
    KI("Kiribati", "KIR", 296),
    KP("Democratic People's Republic of Korea", "PRK", 408),
    KR("Republic of Korea", "KOR", 410),
    KW("Kuwait", "KWT", 414),
    KG("Kyrgyzstan", "KGZ", 417),
    LA("Lao People's Democratic Republic", "LAO", 418),
    LV("Latvia", "LVA", 428),
    LB("Lebanon", "LBN", 422),
    LS("Lesotho", "LSO", 426),
    LR("Liberia", "LBR", 430),
    LY("Libyan Arab Jamahiriya", "LBY", 434),
    LI("Liechtenstein", "LIE", 438),
    LT("Lithuania", "LTU", 440),
    LU("Luxembourg", "LUX", 442),
    MO("Macao", "MAC", 446),
    MK("The Former Yugoslav Republic of Macedonia", "MKD", 807),
    MG("Madagascar", "MDG", 450),
    MW("Malawi", "MWI", 454),
    MY("Malaysia", "MYS", 458),
    MV("Maldives", "MDV", 462),
    ML("Mali", "MLI", 466),
    MT("Malta", "MLT", 470),
    MH("Marshall Islands", "MHL", 584),
    MQ("Martinique", "MTQ", 474),
    MR("Mauritania", "MRT", 478),
    MU("Mauritius", "MUS", 480),
    YT("Mayotte", "MYT", 175),
    MX("Mexico", "MEX", 484),
    FM("Federated States of Micronesia", "FSM", 583),
    MD("Republic of Moldova", "MDA", 498),
    MC("Monaco", "MCO", 492),
    MN("Mongolia", "MNG", 496),
    ME("Montenegro", "MNE", 499),
    MS("Montserrat", "MSR", 500),
    MA("Morocco", "MAR", 504),
    MZ("Mozambique", "MOZ", 508),
    MM("Myanmar", "MMR", 104),
    NA("Namibia", "NAM", 516),
    NR("Nauru", "NRU", 520),
    NP("Nepal", "NPL", 524),
    NL("Netherlands", "NLD", 528),
    NC("New Caledonia", "NCL", 540),
    NZ("New Zealand", "NZL", 554),
    NI("Nicaragua", "NIC", 558),
    NE("Niger", "NER", 562),
    NG("Nigeria", "NGA", 566),
    NU("Niue", "NIU", 570),
    NF("Norfolk Island", "NFK", 574),
    MP("Northern Mariana Islands", "MNP", 580),
    NO("Norway", "NOR", 578),
    PS("Occupied Palestinian Territory", "PSE", 275),
    OM("Oman", "OMN", 512),
    PK("Pakistan", "PAK", 586),
    PW("Palau", "PLW", 585),
    PA("Panama", "PAN", 591),
    PG("Papua New Guinea", "PNG", 598),
    PY("Paraguay", "PRY", 600),
    PE("Peru", "PER", 604),
    PH("Philippines", "PHL", 608),
    PN("Pitcairn", "PCN", 612),
    PL("Poland", "POL", 616),
    PT("Portugal", "PRT", 620),
    PR("Puerto Rico", "PRI", 630),
    QA("Qatar", "QAT", 634),
    RE("Reunion", "REU", 638),
    RO("Romania", "ROU", 642),
    RU("Russian Federation", "RUS", 643),
    RW("Rwanda", "RWA", 646),
    BL("Saint Barthelemy", "534", 703),
    SH("Saint Helena, Ascension and Tristan da Cunha", "SHN", 654),
    KN("Saint Kitts and Nevis", "KNA", 659),
    LC("Saint Lucia", "LCA", 662),
    MF("Saint Martin (French part)", "MAF", 663),
    PM("Saint Pierre and Miquelon", "SPM", 666),
    VC("Saint Vincent and The Grenadines", "VCT", 670),
    WS("Samoa", "WSM", 882),
    SM("San Marino", "SMR", 674),
    ST("Sao Tome and Principe", "STP", 678),
    SA("Saudi Arabia", "SAU", 682),
    SN("Senegal", "SEN", 686),
    RS("Serbia", "SRB", 688),
    SC("Seychelles", "SYC", 690),
    SL("Sierra Leone", "SLE", 694),
    SG("Singapore", "SGP", 702),
    SX("Sint Maarten (Dutch part)", "SXM", 534),
    SK("Slovakia", "SVK", 703),
    SI("Slovenia", "SVN", 705),
    SB("Solomon Islands", "SLB", 90),
    SO("Somalia", "SOM", 706),
    ZA("South Africa", "ZAF", 710),
    GS("South Georgia and the South Sandwich Islands", "SGS", 239),
    SS("South Sudan", "SSD", 728),
    ES("Spain", "ESP", 724),
    LK("Sri Lanka", "LKA", 144),
    SD("Sudan", "SDN", 729),
    SR("Suriname", "SUR", 740),
    SJ("Svalbard and Jan Mayen", "534", 703),
    SZ("Swaziland", "SWZ", 748),
    SE("Sweden", "SWE", 752),
    CH("Switzerland", "CHE", 756),
    SY("Syrian Arab Republic", "SYR", 760),
    TW("Taiwan, Province of China", "TWN", 158),
    TJ("Tajikistan", "TJK", 762),
    TZ("United Republic of Tanzania", "TZA", 834),
    TH("Thailand", "THA", 764),
    TL("Timor-Leste", "TLS", 626),
    TG("Togo", "TGO", 768),
    TK("Tokelau", "TKL", 772),
    TO("Tonga", "TON", 776),
    TT("Trinidad and Tobago", "TTO", 780),
    TN("Tunisia", "TUN", 788),
    TR("Turkey", "TUR", 792),
    TM("Turkmenistan", "TKM", 795),
    TC("Turks and Caicos Islands", "TCA", 796),
    TV("Tuvalu", "TUV", 798),
    UG("Uganda", "UGA", 800),
    UA("Ukraine", "UKR", 804),
    AE("United Arab Emirates", "ARE", 784),
    GB("United Kingdom", "GBR", 826),
    US("United States", "USA", 840),
    UM("United States Minor Outlying Islands", "UMI", 581),
    UY("Uruguay", "URY", 858),
    UZ("Uzbekistan", "UZB", 860),
    VU("Vanuatu", "VUT", 548),
    VE("Bolivarian Republic of Venezuela", "VEN", 862),
    VN("Viet Nam", "VNM", 704),
    VG("British Virgin Islands", "VGB", 92),
    VI("U.S. Virgin Islands", "VIR", 850),
    WF("Wallis and Futuna", "WLF", 876),
    EH("Western Sahara", "ESH", 732),
    YE("Yemen", "YEM", 887),
    ZM("Zambia", "ZMB", 894),
    ZW("Zimbabwe", "ZWE", 716),
;

    private static final Map<String, ISOCountryCode> alpha3CodeMap = new HashMap<String, ISOCountryCode>();
    private static final Map<Integer, ISOCountryCode> numericCodeMap = new HashMap<Integer, ISOCountryCode>();

    private static final PhoneNumberUtil PHONE_NUMBER_UTIL = PhoneNumberUtil.getInstance();
    
    static {
    	
        for (ISOCountryCode next : values()) {
            alpha3CodeMap.put(next.getAlpha3Code(), next);
            numericCodeMap.put(next.getNumericCode(), next);
        }
    }

    private final String countryName;
    private final String alpha3Code;
    private final Integer numericCode;

    private ISOCountryCode(String countryName, String alpha3Code, Integer numericCode) {
        this.countryName = countryName;
        this.alpha3Code = alpha3Code;
        this.numericCode = numericCode;
    }

	/**
	 * Get the country name.
	 * @return The country name.
	 */
	public String getCountryName() {
        return countryName;
    }

	/**
	 * Get the <a href="http://en.wikipedia.org/wiki/ISO_3166-1_alpha-2" >ISO 3166-1 alpha-2</a> code.
	 * @return The <a href="http://en.wikipedia.org/wiki/ISO_3166-1_alpha-2">ISO 3166-1 alpha-2</a> code.
	 */
	public String getAlpha2Code() {
		return name();
	}

	/**
	 * Get the <a href="http://en.wikipedia.org/wiki/ISO_3166-1_alpha-3" >ISO 3166-1 alpha-3</a> code.
	 * @return The <a href="http://en.wikipedia.org/wiki/ISO_3166-1_alpha-3">ISO 3166-1 alpha-3</a> code.
	 */
	public String getAlpha3Code() {
		return alpha3Code;
	}

	/**
	 * Get the <a href="http://en.wikipedia.org/wiki/ISO_3166-1_numericCode" >ISO 3166-1 numericCode</a> code.
	 * @return The <a href="http://en.wikipedia.org/wiki/ISO_3166-1_numericCode">ISO 3166-1 numericCode</a> code.
	 */
	public Integer getNumericCode() {
		return numericCode;
	}
	
	/**
	 * Gets the <a href="http://en.wikipedia.org/wiki/E.164">International Direct Dial</a> prefix for the country
	 * @return
	 */
	public Integer getDiallingRegionCode() {
		return PHONE_NUMBER_UTIL.getCountryCodeForRegion(this.getAlpha2Code());
	}

	/**
	 * Get a CountryCode that corresponds to a given ISO 3166-1 <a href="http://en.wikipedia.org/wiki/ISO_3166-1_alpha-2">alpha-2</a> or 
	 * <a href="http://en.wikipedia.org/wiki/ISO_3166-1_alpha-3">alpha-3</a> code.
	 * @param code An ISO 3166-1 <a href="http://en.wikipedia.org/wiki/ISO_3166-1_alpha-2">alpha-2</a> or 
	 * <a href="http://en.wikipedia.org/wiki/ISO_3166-1_alpha-3">alpha-3</a> code.
	 * @return A ISOCountryCode instance, or null if not found.
	 */
	public static ISOCountryCode getByCode(String code) {
		
		if (code == null) {
			return null;
		}

		if (code.length() == 2) {
			return getByAlpha2Code(code);
		} else if (code.length() == 3) {
			return getByAlpha3Code(code);
		} else {
			return null;
		}
	}

	/**
	 * Get a CountryCode that corresponds to a given ISO 3166-1 <a href="http://en.wikipedia.org/wiki/ISO_3166-1_alpha-2">alpha-2</a> code.
	 * @param code An ISO 3166-1 <a href="http://en.wikipedia.org/wiki/ISO_3166-1_alpha-2">alpha-2</a>.
	 * @return A ISOCountryCode instance, or null if not found.
	 */
	public static ISOCountryCode getByAlpha2Code(String code) {
		try {
			return Enum.valueOf(ISOCountryCode.class, code);
		} catch (IllegalArgumentException e) {
			return null;
		}
	}

	/**
	 * Get a CountryCode that corresponds to a given ISO 3166-1 <a href="http://en.wikipedia.org/wiki/ISO_3166-1_alpha-3">alpha-3</a> code.
	 * @param code An ISO 3166-1 <a href="http://en.wikipedia.org/wiki/ISO_3166-1_alpha-3">alpha-3</a> code.
	 * @return A ISOCountryCode instance, or null if not found.
	 */

	public static ISOCountryCode getByAlpha3Code(String code) {
		return alpha3CodeMap.get(code);
	}

	/**
	 * Get a CountryCode that corresponds to a given <a href="http://en.wikipedia.org/wiki/ISO_3166-1_numericCode">ISO 3166-1
	 * numericCode</a> code.
	 * @param code An <a href="http://en.wikipedia.org/wiki/ISO_3166-1_numericCode">ISO 3166-1 numericCode</a> code.
	 * @return A CountryCode instance, or null if not found.
	 */
	public static ISOCountryCode getByNumericCode(Integer code) {
		return numericCodeMap.get(code);
	}

	/**
	 * Get a CountryCode that corresponds to a given <a href="http://en.wikipedia.org/wiki/E.164">Dialling Region</a> code.
	 * @param code An <a href="http://en.wikipedia.org/wiki/E.164">E.164 Dialling Region Code</a> code.
	 * @return A CountryCode instance, or null if not found.
	 */
	public static ISOCountryCode getByDiallingRegionCode(Integer code) {
		return getByCode(PHONE_NUMBER_UTIL.getRegionCodeForCountryCode(code));
	}
}